#!/usr/bin/env python
# Copyright 2021, TATA Consultancy Services. All rights reserved.

# Edit this script to add your team's training code.
# Some functions are *required*, but you can edit most parts of the required functions, remove non-required functions, and add your own functions.

################################################################################
#
# Imported functions and variables
#
################################################################################
import tensorflow

seed_value = 42
seed = 42
tensorflow.random.set_seed(seed)
tensorflow.compat.v1.set_random_seed(seed_value)

# Import functions. These functions are not required. You can change or remove them.
from helper_code import *
import numpy as np, os, sys, joblib
from sklearn.impute import SimpleImputer
from sklearn.ensemble import RandomForestClassifier

#---Varsha Edit1----------------------------------#
import time
import random
import pandas as pd
import pickle
import joblib
from scipy.io import loadmat
from sklearn.model_selection import train_test_split

from Parameter import Parameter_dictionary
from get_12ECG_features import get_12ECG_features
from CNNet import CNNet

import tensorflow.keras.models
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint

#test libraries
from tensorflow.keras.models import load_model
from sklearn.cluster import KMeans
from sklearn.preprocessing import MinMaxScaler

seed_value = 20
# 1. Set `PYTHONHASHSEED` environment variable at a fixed value
os.environ['PYTHONHASHSEED']=str(seed_value)
# 2. Set `python` built-in pseudo-random generator at a fixed value
random.seed(seed_value)
# 3. Set `numpy` pseudo-random generator at a fixed value
np.random.seed(seed_value)
#---------------------------------------------------------#

#Define the Challenge lead sets. These variables are not required. You can change or remove them.
twelve_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF', 'V1', 'V2', 'V3', 'V4', 'V5', 'V6')
six_leads = ('I', 'II', 'III', 'aVR', 'aVL', 'aVF')
four_leads = ('I', 'II', 'III', 'V2')
three_leads = ('I', 'II', 'V2')
two_leads = ('I', 'II')

lead_sets = (twelve_leads, six_leads, four_leads, three_leads, two_leads)

def getIndices(leads):
    if leads == twelve_leads :
        Indices = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    elif leads == six_leads:
        Indices = [0, 1, 2, 3, 4, 5]
    elif leads == four_leads:
        Indices = [0, 1, 2, 7]
    elif leads == three_leads:
        Indices = [0, 1, 7]
    elif leads == two_leads:
        Indices = [0, 1]
    return Indices
################################################################################
#
# Training model function
#
################################################################################

# Train your model. This function is *required*. You should edit this function to add your code, but do *not* change the arguments of this function.
def training_code(data_directory, model_directory):
    #--------- Varsha Edit2---------------#
    Param = Parameter_dictionary()
    UniformTimelength = Param.get("TimeLength")
    Sampling_Rate_uniform = Param.get("Uniform_Sampling_Rate")
    SliceLength = Param.get("SliceLength")

    ColumnCount = UniformTimelength * Sampling_Rate_uniform
    nrw = Param.get("nrw")# check
    ReshapeInputData = Param.get("ReshapeInputData")
    MinMaxNor = Param.get("MinMaxNormalization")
    CNN_LSTM = Param.get("CNN_LSTM")
    CNNet_val = Param.get("CNNet")
    OnlyScoringClass = Param.get("TrainOnlyScoringClasses")

    MultiLabel_2_multiClass = Param.get("MultiLabel_2_multiClass")
    loss_function = 'binary_crossentropy'
    if MultiLabel_2_multiClass == 1:
        loss_function = 'categorical_crossentropy'
    #-------------------------------------#

    # Find header and recording files.
    print('Finding header and recording files...')

    header_files, recording_files = find_challenge_files(data_directory) # func from helper code
    num_recordings = len(recording_files)

    if not num_recordings:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    if not os.path.isdir(model_directory):
        os.mkdir(model_directory)

    # Extract the classes from the dataset.
    print('Extracting classes...')

    classes = set()
    for header_file in header_files:
        header = load_header(header_file)
        classes |= set(get_labels(header))
    if all(is_integer(x) for x in classes):
        classes = sorted(classes, key=lambda x: int(x)) # Sort classes numerically if numbers.
    else:
        classes = sorted(classes) # Sort classes alphanumerically if not numbers.
    num_classes = len(classes)

    #--------- Varsha Edit3---------------#
    print('Number of files read:', num_recordings)
    print('Number of classes present:', num_classes)

    recordings = list()
    headers = list()
    labels = np.zeros((num_recordings, num_classes), dtype=np.bool)
    for i,header_file in enumerate(header_files):
        #print(header_file)
        #print(i)
        recording = load_recording(recording_files[i])
        recording = np.nan_to_num(recording)
        recordings.append(recording)

        header = load_header(header_file)
        headers.append(header)
        current_labels = get_labels(header)  # func from helper code
        for label in current_labels:
            if label in classes:
                j = classes.index(label)
                labels[i, j] = 1
    # Convert boolean labels to binary labels
    labels = labels*1
    # ----------------------------------------------#

    # --------- Varsha Edit4--------------------#
    if OnlyScoringClass == 1:
        recordings, labels, classes = OnlyScoringClassInstances(recordings, labels, classes)
        print('Number of files remaining  after retaining labels containing scoring classes:', len(labels))
        print('Number of classes remaining  after retaining labels containing scoring classes:', len(classes))

    print('Training and validation data preprocessing and feature generation...')
    # ------------------------------------------#

    # Extract the features and labels from the dataset.
    print('Extracting features and labels...')

    # --------- Varsha Edit6--------------------#
    features = list()
    DNN_ready_data = list()
    Labels_repeated_master = list()

    for i in range(len(labels)):
        #print(i)
        #print(f)
        #if i == 4053:
        recording = recordings[i]
        #print(recording.shape)
        header = headers[i]
        CurrentLabel = labels[i, :]
        tmp, generated_features = get_12ECG_features(recording, header, Param)
        DataSliced, LabelRepeat, generated_features_repeated = Data_slicing(tmp, Param, CurrentLabel,
                                                                            generated_features)
        DNN_ready_data.extend(DataSliced)
        features.extend(generated_features_repeated)
        Labels_repeated_master.extend(LabelRepeat)

    features = np.array(features)
    features = np.nan_to_num(features)
    DNN_ready_data = np.array(DNN_ready_data)
    Labels_repeated_master = np.array(Labels_repeated_master)
    labels = Labels_repeated_master
    ColumnCount = DNN_ready_data.shape[2]

    indices = []
    for i in range(len(DNN_ready_data)):
        indices.append(i)
    indices = np.array(indices)
    X_train, X_valid, y_train, y_valid, train_indices, valid_indices = train_test_split(DNN_ready_data, labels, indices,
                                                                                        test_size=0.2, random_state=42)
    print('Number of training files:', len(y_train))
    print('Number of validating files:', len(y_valid))
    X_train_HCF = features[train_indices, :]
    X_valid_HCF = features[valid_indices, :]

    # Normalized HCF
    Min_X_train_HCF = np.min(X_train_HCF, axis=0)
    Max_X_train_HCF = np.std(X_train_HCF, axis=0)
    Range = Max_X_train_HCF - Min_X_train_HCF
    zeroInd = np.where(Range == 0)[0]
    if len(zeroInd) != 0:
        Range[zeroInd] = 1

    X_train_HCF = (X_train_HCF - Min_X_train_HCF) / Range
    X_valid_HCF = (X_valid_HCF - Min_X_train_HCF) / Range
    #

    ### Weight computation
    train_class_distribution = np.sum(y_train, axis=0)
    train_class_distribution_scaled = train_class_distribution / np.sum(train_class_distribution)
    for i in range(len(train_class_distribution_scaled)):
        train_class_distribution_scaled[i] = np.round(1 / train_class_distribution_scaled[i], 2)
    print('Data dimension:', X_train.shape[1], X_train.shape[2])

    MaxVal_Per_Channel = [0] * nrw
    MaxVal_Per_Channel = np.array(MaxVal_Per_Channel)
    MinVal_Per_Channel = [0] * nrw
    MinVal_Per_Channel = np.array(MinVal_Per_Channel)
    if MinMaxNor == 1:
        # Minmax normalization
        print('Min-Max Normalization applied')

        for i in range(nrw):
            ith_Channel_train_data = X_train[:, i, :]
            ith_Channel_train_data_max = ith_Channel_train_data.max(axis=1)
            ith_Channel_train_data_min = ith_Channel_train_data.min(axis=1)
            MaxVal_Per_Channel[i] = np.percentile(ith_Channel_train_data_max, 90)
            MinVal_Per_Channel[i] = np.percentile(ith_Channel_train_data_min, 10)
        MaxVal_Per_Channel = MaxVal_Per_Channel.reshape(len(MaxVal_Per_Channel), 1)
        MinVal_Per_Channel = MinVal_Per_Channel.reshape(len(MinVal_Per_Channel), 1)
        Range_per_Channel = MaxVal_Per_Channel - MinVal_Per_Channel
        if 0 in Range_per_Channel:
            Zero_Range_loc = np.where(Range_per_Channel == 0)
            Range_per_Channel[Zero_Range_loc] = 1
        for i in range(nrw):
            ith_Channel_train_data = X_train[:, i, :]
            ith_Channel_train_data = (ith_Channel_train_data - MinVal_Per_Channel[i]) / Range_per_Channel[i]
            X_train[:, i, :] = ith_Channel_train_data
        for i in range(nrw):
            ith_Channel_valid_data = X_valid[:, i, :]
            ith_Channel_valid_data = (ith_Channel_valid_data - MinVal_Per_Channel[i]) / Range_per_Channel[i]
            X_valid[:, i, :] = ith_Channel_valid_data

    if ReshapeInputData == 1:
        X_train = X_train.reshape(len(X_train), int(nrw * 8), int((Sampling_Rate_uniform * UniformTimelength) / 8))
        X_valid = X_valid.reshape(len(X_valid), int(nrw * 8), int((Sampling_Rate_uniform * UniformTimelength) / 8))
        ColumnCount = int((Sampling_Rate_uniform * UniformTimelength) / 8)
        nrw = int(nrw * 8)


    #--------Varsha Edit9--------------------------------------#
    # Train a model for each lead set.
    for leads in lead_sets:
        Indices= getIndices(leads)
        print('Training model for {}-lead set: {}...'.format(len(leads), ', '.join(leads)))
        # Extract the features for the model.
        X_train1 = X_train[:,Indices]
        X_valid1 = X_valid[:,Indices]

        #print(X_train[:, leads].shape)
        num_lead = X_train1.shape[1]
        # Train the model.
        train(X_train1, y_train, X_valid1, y_valid, ColumnCount,num_lead, classes,Param, model_directory, MaxVal_Per_Channel,
              MinVal_Per_Channel, CNNet_val, CNN_LSTM, loss_function, X_train_HCF, X_valid_HCF, Min_X_train_HCF,
              Max_X_train_HCF,leads)

################################################################################
#
# Running trained model function
#
################################################################################

#-Varsha:----- This is our run_12ECG code use to generate test features and to fit model on generated features--#
# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the arguments of this function.
def run_model(model, header, recording):
    #-----------------Varsha Edit11----------------------#
    leads = get_leads(header)

    # Use your classifier here to obtain a label and score for each class.
    f_out = 'address.pkl'
    filename = os.path.join(os.getcwd(), f_out)

    pkl_file = open(filename, 'rb')
    mydict2 = pickle.load(pkl_file)
    pkl_file.close()

    filename = mydict2 + "/final_model_parameters_"+get_model_filename(leads)+".pickle"

    pkl_file = open(filename, 'rb')
    loaded_model_parameters = pickle.load(pkl_file)
    pkl_file.close()

    classes = loaded_model_parameters['classes']
    Param = loaded_model_parameters['Parameters']
    MaxVal_Per_Channel = loaded_model_parameters['MinMax_nor_max']
    MinVal_Per_Channel = loaded_model_parameters['MinMax_nor_min']
    Min_X_train_HCF = loaded_model_parameters['Min_train_HCF']
    Max_X_train_HCF = loaded_model_parameters['Max_train_HCF']
    ReshapeInputData = Param.get("ReshapeInputData")
    UniformTimelength = Param.get("TimeLength")
    Sampling_Rate_uniform = Param.get("Uniform_Sampling_Rate")
    SliceLength = Param.get("SliceLength")
    ColumnCount = SliceLength * Sampling_Rate_uniform
    nrw = len(leads)#Param.get("nrw") # Varsha Edit
    MinMaxNor = Param.get("MinMaxNormalization")
    CNN_LSTM = Param.get("CNN_LSTM")
    CNNet_val = Param.get("CNNet")

    DNN_ready_data, generated_features = get_12ECG_features(recording, header, Param)
    generated_features = np.nan_to_num(generated_features)
    DNN_ready_data = np.nan_to_num(DNN_ready_data)

    DNN_ready_data, generated_features = Data_slicing_test(DNN_ready_data, Param, generated_features)

    # Normalized HCF
    Range = Max_X_train_HCF - Min_X_train_HCF
    zeroInd = np.where(Range == 0)[0]
    if len(zeroInd) != 0:
        Range[zeroInd] = 1

    generated_features = (generated_features - Min_X_train_HCF) / Range

    if MinMaxNor == 1:
        Range_per_Channel = MaxVal_Per_Channel - MinVal_Per_Channel
        DNN_ready_data = (DNN_ready_data - MinVal_Per_Channel) / Range_per_Channel
    if ReshapeInputData == 1:
        DNN_ready_data = DNN_ready_data.reshape(int(nrw * 8), int((Sampling_Rate_uniform * UniformTimelength) / 8))
        ColumnCount = int((Sampling_Rate_uniform * UniformTimelength) / 8)
        nrw = int(nrw * 8)
    if CNNet_val == 1:
        DNN_ready_data = DNN_ready_data.reshape(DNN_ready_data.shape[0], nrw, ColumnCount,
                                                1)  # (1, nrw, batchS) # Remove 4th dimension for Chinese model
    if CNN_LSTM == 1:
        DNN_ready_data = DNN_ready_data.transpose(1, 0)
        n_steps, n_length = 4, 150
        n_features = nrw
        DNN_ready_data = DNN_ready_data.reshape((1, n_steps, n_length, n_features))

    current_score_list = list()
    for i in range(int(DNN_ready_data.shape[0])):
        DNN_ready_dataR = DNN_ready_data[:, 1, :, :]
        current_score = model.predict([DNN_ready_data, DNN_ready_dataR, generated_features])
        current_score = np.around(current_score, decimals=3)
        current_score_list.extend(current_score)
    current_score = np.mean(current_score, axis=0)
    current_score = np.transpose(np.reshape(current_score, [-1, 1]))
    current_label = Pred_prob2Pred_class(current_score)
    current_label = current_label.astype(int)#labels
    current_score = np.asarray(current_score)
    current_score = current_score[0] #probabilities
    #-----------------------------------------------------#
    return classes, current_label, current_score

#------------Varsha Edit10------------------#
def Pred_prob2Pred_class(Predicted_scores):
    Sorted_Predicted_scores=np.transpose(Predicted_scores)
    kmeans = KMeans(n_clusters=2, random_state=45).fit(Sorted_Predicted_scores)
    Actlabel = kmeans.labels_
    centers = kmeans.cluster_centers_
    c1 = centers[0][0]
    c2 = centers[1][0]
    #print(c1, c2)
    #print(Actlabel)
    if c1 > c2:
        id1 = np.where(Actlabel == 1)[0]
        id2 = np.where(Actlabel == 0)[0]
        Actlabel[id1] = 0
        Actlabel[id2] = 1
    # Predicted_Class_indices = np.where(Predicted_scores>0.01)
    # Predicted_Class_indices = np.array(Predicted_Class_indices)
    # Predicted_Classes = np.zeros((len(Predicted_scores[0]),), dtype=int)
    # Predicted_Classes[Predicted_Class_indices] = 1
    return Actlabel
#-------------------------------------------#
################################################################################
#
# File I/O functions
#
################################################################################
# Load a trained model. This function is *required*. You should edit this function to add your code, but do *not* change the arguments of this function.

def load_model(model_directory, leads):
    #--------Varsha Edit9-------------------------#
    # load the model from disk
    f_out = 'model_'+get_model_filename(leads)+'.hdf5'
    filename = os.path.join(model_directory, f_out)
    loaded_model = tensorflow.keras.models.load_model(filename)

    Model_address = model_directory
    present_working_directory = os.getcwd()
    f_out = 'address.pkl'
    filename = os.path.join(present_working_directory, f_out)
    output = open(filename, 'wb')
    pickle.dump(Model_address, output)
    output.close()

    return loaded_model

# Define the filename(s) for the trained models. This function is not required. You can change or remove it.
def get_model_filename(leads):
    sorted_leads = sort_leads(leads)
    return '-'.join(sorted_leads)

#----------- Varsha Edit5------------------------------#
def OnlyScoringClassInstances(data,labels,classes):
    data_screened = list()
    labels_screened = list()
    Scored_Labels = np.loadtxt("scored_labels.csv", dtype=str) # 27 Labels
    scoring_labels = Scored_Labels.tolist()
    scoring_labels = set(scoring_labels).intersection(classes)
    scoring_labels = list(scoring_labels)
    ScoringClass_mapping = np.zeros(len(scoring_labels))
    for i in range(len(scoring_labels)):
        ScoringClass_mapping[i] = classes.index(scoring_labels[i])
    ScoringClass_mapping_int = ScoringClass_mapping.astype(int)
    for i in range(len(labels)):
        Current_label = labels[i,:]
        Current_data = data[i]
        ScoringClass_loc_current_label = Current_label[ScoringClass_mapping_int]
        if (sum(ScoringClass_loc_current_label)>=1):
            data_screened.append(Current_data)
            labels_screened.append(Current_label)
    labels = np.array(labels_screened)
    labels = labels[:,ScoringClass_mapping_int]
    # data_screened = np.array(data_screened)
    return data_screened, labels,scoring_labels


def Limit_unique_classes(data, labels):
    unique_labels = np.unique(labels, axis=0)
    Count_Classes_retained = int(np.log(len(labels)) + np.sqrt(len(labels)))
    Trimmed_Labels = list()
    Trimmed_data = list()

    Unique_labels_count = list()
    Loc_iden_label_instances = [0]
    Loc_iden_label_instances_store = list()

    for unique_label in unique_labels:
        Loc_iden_label_instances = np.where(np.all(labels == np.transpose(unique_label), axis=1))
        Loc_iden_label_instances_store.append(Loc_iden_label_instances)
        Unique_labels_count.append(len(Loc_iden_label_instances[0]))

    SortedIndices = np.argsort(np.array(Unique_labels_count))
    SortedIndices = SortedIndices[::-1]
    SortedIndices = SortedIndices[0:Count_Classes_retained]
    for i in SortedIndices:
        locations = Loc_iden_label_instances_store[i]

        temp_label = np.array(labels[locations])
        Trimmed_Labels.extend(temp_label)
        for l in locations[0]:
            Trimmed_data.append(data[l])
    Trimmed_Labels = np.array(Trimmed_Labels)
    Trimmed_data = np.array(Trimmed_data)
    return Trimmed_data, Trimmed_Labels


# def Data_slicing(data, Param, CurrentLabel, generated_features):
#     DataSliced = list()
#     LabelRepeat = list()
#     generated_features_repeated = list()
#     SlicedDataLength = Param.get("SliceLength")
#     Target_sampling_rate = Param.get("Uniform_Sampling_Rate")
#     SliceWidth = SlicedDataLength * Target_sampling_rate
#     dataLen = data.shape[1]
#     SliceCount = int(np.floor(dataLen / SliceWidth))
#     data_extention = 0
#     while (dataLen < SliceWidth):
#         data = np.repeat(data, 2, axis=1)
#         dataLen = data.shape[1]
#         data_extention = 1
#     if data_extention == 1:
#         data = data[:, 0:SliceWidth]
#     SliceCount = int(np.floor(dataLen / SliceWidth))
#     # print('SliceWidth,dataLen,SliceCount:',SliceWidth,dataLen,SliceCount)
#     for i in range(SliceCount):
#         DataSliced.append(data[:, 0 + i * SliceWidth:((i + 1) * SliceWidth)])
#         LabelRepeat.append(CurrentLabel)
#         generated_features_repeated.append(generated_features)
#     DataSliced = np.array(DataSliced)
#     LabelRepeat = np.array(LabelRepeat)
#     generated_features_repeated = np.array(generated_features_repeated)
#     return DataSliced, LabelRepeat, generated_features_repeated


def Data_slicing(data, Param, CurrentLabel, generated_features):
    bias = int(0)
    DataSliced = list()
    LabelRepeat = list()
    generated_features_repeated = list()
    SlicedDataLength = Param.get("SliceLength")
    Target_sampling_rate = Param.get("Uniform_Sampling_Rate")
    Overlap_percent = Param.get("Overlap_percent")
    TimeForOverlap_InSeconds = Param.get("TimeForOverlap_InSeconds")
    SliceWidth = SlicedDataLength * Target_sampling_rate
    dataLen = data.shape[1]
    Channel_count = data.shape[0]
    SliceCount = int(np.floor(dataLen / SliceWidth))
    data_extention = 0
    while (dataLen < SliceWidth):
        data_len_deficit = SliceWidth - dataLen
        Right_zero_padding = int(np.ceil(data_len_deficit / 2))
        Zero_vect_R = np.zeros((1, Right_zero_padding), dtype=int)
        Zero_vect_R = np.repeat(Zero_vect_R, Channel_count, axis=0)
        Left_zero_padding = data_len_deficit - Right_zero_padding
        Zero_vect_L = np.zeros((1, Left_zero_padding), dtype=int)
        Zero_vect_L = np.repeat(Zero_vect_L, Channel_count, axis=0)
        data = np.concatenate((Zero_vect_L, data, Zero_vect_R), axis=1)

        # data = np.repeat(data, 2, axis=1)
        dataLen = data.shape[1]
        data_extention = 1
    if data_extention == 1:
        data = data[:, 0:SliceWidth]

    # print('SliceWidth,dataLen,SliceCount:',SliceWidth,dataLen,SliceCount)

    # sliding window with 50% overlap only when data is less than equals to 30s
    # print(data.shape[1])
    Overlap_sample = int(Overlap_percent * SliceWidth)

    sample_thres_for_overlap = TimeForOverlap_InSeconds * Target_sampling_rate
    if data.shape[1] <= sample_thres_for_overlap:
        SliceCount = int(np.floor(dataLen / Overlap_sample)) - 1
    else:
        SliceCount = int(np.floor(dataLen / SliceWidth))
    if data.shape[1] > SliceWidth+11*Overlap_sample: #for signals >12 windows, the signal is trimmed to contain only 12 windows
        # SliceCount = int(np.floor(SliceWidth * 12 / Overlap_sample)) - 1
        SliceCount = 12
        bias = int(np.floor(dataLen / 2) - (SliceWidth+11*Overlap_sample)/2)
    # print(data.shape[1])
    # print(SliceCount)
    # print(data.shape[1])
    for i in range(SliceCount):
        # if data.shape[1] <= sample_thres_for_overlap:
        a = bias + i * Overlap_sample
        b = bias + ((i * Overlap_sample) + SliceWidth)
        # else:
        #     a = 0 + i * SliceWidth
        #     b = ((i + 1) * SliceWidth)
        # if data.shape[1] > 12*SliceWidth:
        #     a = 0 + i * SliceWidth
        #     b = ((i + 1) * SliceWidth)
        # print(a,b)
        if b <= data.shape[1]:  # (((SliceCount - 1) * Overlap_sample) + SliceWidth) <= data.shape[1]:
            DataSliced.append(data[:, a:b])
            LabelRepeat.append(CurrentLabel)
            generated_features_repeated.append(generated_features)
            # print(a, b)

    DataSliced = np.array(DataSliced)
    LabelRepeat = np.array(LabelRepeat)
    generated_features_repeated = np.array(generated_features_repeated)
    # if SliceCount >= 2:
    #     for i in range(SliceCount-1):
    #         DataSliced.append(data[:,0+i*SliceWidth:((i+1)*SliceWidth)])
    #         LabelRepeat.append(CurrentLabel)
    #     DataSliced.append(data[:,dataLen-SliceWidth:dataLen])
    #     LabelRepeat.append(CurrentLabel)
    # elif SliceCount == 1:
    #     DataSliced.append(data)
    #     LabelRepeat.append(CurrentLabel)
    # DataSliced = np.array(DataSliced)
    # LabelRepeat = np.array(LabelRepeat)

    return DataSliced, LabelRepeat, generated_features_repeated

# def Data_slicing_test(data, Param, generated_features):
#     DataSliced = list()
#     generated_features_repeated = list()
#     SlicedDataLength = Param.get("SliceLength")
#     Target_sampling_rate = Param.get("Uniform_Sampling_Rate")
#     SliceWidth = SlicedDataLength * Target_sampling_rate
#     dataLen = data.shape[1]
#     SliceCount = int(np.floor(dataLen / SliceWidth))
#     data_extention = 0
#     while (dataLen < SliceWidth):
#         data = np.repeat(data, 2, axis=1)
#         data = data[:, 0:SliceWidth - 1]
#         dataLen = data.shape[1]
#         data_extention = 1
#     if data_extention == 1:
#         data = data[:, 0:SliceWidth]
#     SliceCount = int(np.floor(dataLen / SliceWidth))
#     # print('SliceWidth,dataLen,SliceCount:',SliceWidth,dataLen,SliceCount)
#     for i in range(SliceCount):
#         # print(i)
#         DataSliced.append(data[:, 0 + i * SliceWidth:((i + 1) * SliceWidth)])
#         generated_features_repeated.append(generated_features)
#     DataSliced = np.array(DataSliced)
#     generated_features_repeated = np.array(generated_features_repeated)
#     return DataSliced, generated_features_repeated
def Data_slicing_test(data, Param, generated_features):
    bias = int(0)
    DataSliced = list()
    generated_features_repeated = list()
    SlicedDataLength = Param.get("SliceLength")
    Target_sampling_rate = Param.get("Uniform_Sampling_Rate")
    Overlap_percent = Param.get("Overlap_percent")
    TimeForOverlap_InSeconds = Param.get("TimeForOverlap_InSeconds")
    SliceWidth = SlicedDataLength * Target_sampling_rate
    dataLen = data.shape[1]
    Channel_count = data.shape[0]
    SliceCount = int(np.floor(dataLen / SliceWidth))
    data_extention = 0
    while (dataLen < SliceWidth):
        data_len_deficit = SliceWidth - dataLen
        Right_zero_padding = int(np.ceil(data_len_deficit / 2))
        Zero_vect_R = np.zeros((1, Right_zero_padding), dtype=int)
        Zero_vect_R = np.repeat(Zero_vect_R, Channel_count, axis=0)
        Left_zero_padding = data_len_deficit - Right_zero_padding
        Zero_vect_L = np.zeros((1, Left_zero_padding), dtype=int)
        Zero_vect_L = np.repeat(Zero_vect_L, Channel_count, axis=0)
        data = np.concatenate((Zero_vect_L, data, Zero_vect_R), axis=1)

        # data = np.repeat(data, 2, axis=1)
        dataLen = data.shape[1]
        data_extention = 1
    if data_extention == 1:
        data = data[:, 0:SliceWidth]

    # print('SliceWidth,dataLen,SliceCount:',SliceWidth,dataLen,SliceCount)

    # sliding window with 50% overlap only when data is less than equals to 30s
    # print(data.shape[1])
    Overlap_sample = int(Overlap_percent * SliceWidth)

    sample_thres_for_overlap = TimeForOverlap_InSeconds * Target_sampling_rate
    if data.shape[1] <= sample_thres_for_overlap:
        SliceCount = int(np.floor(dataLen / Overlap_sample)) - 1
    else:
        SliceCount = int(np.floor(dataLen / SliceWidth))
    if data.shape[1] > SliceWidth+11*Overlap_sample: #for signals >12 windows, the signal is trimmed to contain only 12 windows
        # SliceCount = int(np.floor(SliceWidth * 12 / Overlap_sample)) - 1
        SliceCount = 12
        bias = int(np.floor(dataLen / 2) - (SliceWidth+11*Overlap_sample)/2)
    # print(data.shape[1])
    # print(SliceCount)
    # print(data.shape[1])
    for i in range(SliceCount):
        # if data.shape[1] <= sample_thres_for_overlap:
        a = bias + i * Overlap_sample
        b = bias + ((i * Overlap_sample) + SliceWidth)
        # else:
        #     a = 0 + i * SliceWidth
        #     b = ((i + 1) * SliceWidth)
        # if data.shape[1] > 12*SliceWidth:
        #     a = 0 + i * SliceWidth
        #     b = ((i + 1) * SliceWidth)
        # print(a,b)
        if b <= data.shape[1]:  # (((SliceCount - 1) * Overlap_sample) + SliceWidth) <= data.shape[1]:
            DataSliced.append(data[:, a:b])
            # LabelRepeat.append(CurrentLabel)
            generated_features_repeated.append(generated_features)
            # print(a, b)

    DataSliced = np.array(DataSliced)
    generated_features_repeated = np.array(generated_features_repeated)
    return DataSliced, generated_features_repeated
#--------- Varsha Edit8---------------------------------#

def train(X_train, y_train, X_valid, y_valid, ncl, num_lead, considered_classes, Param, model_directory, MaxVal_Per_Channel,
          MinVal_Per_Channel, CNNet_val, CNN_LSTM, loss_function, X_train_HCF, X_valid_HCF, Min_X_train_HCF,
          Max_X_train_HCF,leads):
    num_classes = len(considered_classes)

    if CNNet_val == 1:
        X_train = X_train.reshape(X_train.shape[0], num_lead, ncl,
                                  1)  # (Data, num_lead, ncl) # Remove 4th dimension for Chinese model
        X_valid = X_valid.reshape(X_valid.shape[0], num_lead, ncl,
                                  1)  # (Data, num_lead, ncl) # Remove 4th dimension for Chinese model
        model = CNNet(num_lead, ncl, num_classes)
    if CNN_LSTM == 1:
        X_train = X_train.transpose(0, 2, 1)
        X_valid = X_valid.transpose(0, 2, 1)
        n_steps, n_length = 4, 150
        n_features = num_lead
        X_train = X_train.reshape((X_train.shape[0], n_steps, n_length, n_features))
        X_valid = X_valid.reshape((X_valid.shape[0], n_steps, n_length, n_features))
        # Load Model
        model = Conv_LSTM_Net(n_length, n_features, num_classes)

    from tensorflow.keras.optimizers import Adam
    import json
    params = json.load(open("config_resnet.json"))
    optimizer = Adam(
        lr=params["learning_rate"],
        clipnorm=params.get("clipnorm", 1))

    model.compile(loss=loss_function,
                  optimizer=optimizer,
                  metrics=['accuracy'])

    model.summary()

    filename = model_directory + "/model_"+get_model_filename(leads)+".hdf5"
    checkpoint = ModelCheckpoint(filename,
                                 monitor='val_loss', verbose=1, save_best_only=True, save_weights_only=False,
                                 mode='auto', period=1)

    early = EarlyStopping(monitor='val_loss', min_delta=0, patience=10, verbose=1, mode='auto')

    #====RESNET Data=====================#
    X_trainR = X_train[:, 1, :, :]
    # X_trainR = X_trainR.reshape(140,1,500)
    X_validR = X_valid[:, 1, :, :]
    # X_validR = X_validR.reshape(36, 1, 500)
    #========================================#
    # print("X_train:", X_train.shape)
    # print("X_trainR:", X_trainR.shape)
    # print("X_valid:",X_valid.shape)
    # print("X_validR:", X_validR.shape)

    reduce_lr = tensorflow.keras.callbacks.ReduceLROnPlateau(
        factor=0.1,
        patience=2,
        min_lr=params["learning_rate"] * 0.001)

    history = model.fit([X_train, X_trainR, X_train_HCF], y_train,
                        validation_data=([X_valid, X_validR, X_valid_HCF], y_valid),
                        batch_size=32, epochs=200, callbacks=[checkpoint, reduce_lr, early])

    # import matplotlib.pyplot as pyplot
    # fig1 = pyplot.gcf()
    # # pyplot.yscale('log')
    # pyplot.title('Learning Curves')
    # pyplot.xlabel('Epoch')
    # pyplot.ylabel('Loss')
    # pyplot.plot(history.history['loss'], label='train')
    # pyplot.plot(history.history['val_loss'], label='val')
    # pyplot.legend()
    # fig1.savefig(model_directory + '/train_vs_eval_loss_'+get_model_filename(leads)+'.png', dpi=300)
    # pyplot.close()
    #
    # fig2 = pyplot.gcf()
    # # pyplot.yscale('log')
    # pyplot.title('Learning Curves')
    # pyplot.xlabel('Epoch')
    # pyplot.ylabel('Accuracy')
    # pyplot.plot(history.history['accuracy'], label='train')
    # pyplot.plot(history.history['val_accuracy'], label='val')
    # pyplot.legend()
    # fig2.savefig(model_directory + '/train_vs_eval_accuracy_'+get_model_filename(leads)+'.png', dpi=300)
    # pyplot.close()
    #
    # # Save History
    # pd.DataFrame(history.history).to_csv(path_or_buf=model_directory + '/History_'+get_model_filename(leads)+'.csv')

    # Save model.
    print('Saving model parameters...')
    model_parameters = {'classes': considered_classes, 'Parameters': Param, 'MinMax_nor_max': MaxVal_Per_Channel,
                        'MinMax_nor_min': MinVal_Per_Channel, 'Min_train_HCF': Min_X_train_HCF,
                        'Max_train_HCF': Max_X_train_HCF}

    filename = model_directory + "/final_model_parameters_"+get_model_filename(leads)+".pickle"

    output = open(filename, 'wb')
    pickle.dump(model_parameters, output)
    output.close()
    print('Model and parameters saved...')
    #-------------------------------------------------------------------------------------------------#



